/** 
* Calculates the 1D index of the 3D array with the given sizes.
*/
int indexOf(int x, int y, int z, int xSize, int ySize, int zSize)
{
    int index = z * (xSize*ySize) + y*xSize + x;

    return index;
} //end of indexOf function

/**
* intersect ray with a box
* http://www.siggraph.org/education/materials/HyperGraph/raytrace/rtinter3.htm
*/

int intersectBox(float4 r_o, float4 r_d, float4 boxmin, float4 boxmax, float *tnear, float *tfar)
{
    // compute intersection of ray with all six bbox planes
    float4 invR = (float4)(1.0f,1.0f,1.0f,1.0f) / r_d;
    float4 tbot = invR * (boxmin - r_o);
    float4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    float4 tmin = min(ttop, tbot);
    float4 tmax = max(ttop, tbot);

    // find the largest tmin and the smallest tmax
    float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

	*tnear = largest_tmin;
	*tfar = smallest_tmax;

	return smallest_tmax > largest_tmin;
} //end of intersectBox function

/**
* Returns the int RGBA value for the given float4
*/
uint rgbaFloatToInt(float4 rgba)
{
    rgba.x = clamp(rgba.x,0.0f,1.0f);  
    rgba.y = clamp(rgba.y,0.0f,1.0f);  
    rgba.z = clamp(rgba.z,0.0f,1.0f);  
    rgba.w = clamp(rgba.w,0.0f,1.0f);  
    return ((uint)(rgba.w*255.0f)<<24) | ((uint)(rgba.z*255.0f)<<16) | ((uint)(rgba.y*255.0f)<<8) | (uint)(rgba.x*255.0f);
} //end of rgbaFloatToInt function

/**
* Main kernel function
*/
__kernel void sampleKernel(int imageWidth, int imageHeight, int gridWidth, int gridHeight, int gridDepth, __global float *matrix, __global float *volume, __global uint *pixels)
{
    uint x = get_global_id(0);
    uint y = get_global_id(1);

    //Convert the indices to the range -1 -> + 1;
    float u = (x / (float) imageWidth)*2.0f-1.0f;
    float v = (y / (float) imageHeight)*2.0f-1.0f;
    
    float4 boxMin = (float4)(-1.0f, -1.0f, -1.0f,1.0f);
    float4 boxMax = (float4)(1.0f, 1.0f, 1.0f,1.0f);
    
    // calculate eye ray in world space
    float4 eyeRay_o = 0;
    float4 eyeRay_d = 0;
    
    eyeRay_o = (float4)(matrix[3], matrix[7], matrix[11], 1.0f);  
    //eyeRay_o = (float4)(u, v, +1.f, 1.f);  
    
    float4 temp = normalize(((float4)(u, v, -2.0f,0.0f)));
    eyeRay_d.x = dot(temp, ((float4)(matrix[0],matrix[1],matrix[2],matrix[3])));
    eyeRay_d.y = dot(temp, ((float4)(matrix[4],matrix[5],matrix[6],matrix[7])));
    eyeRay_d.z = dot(temp, ((float4)(matrix[8],matrix[9],matrix[10],matrix[11])));
    eyeRay_d.w = 0.0f;
    


//    eyeRay_o = (float4)(u,v,1,1.0f);
    //eyeRay_d = (float4)(0,0,-1.0f,1.0f);
    
     // find intersection with box
    float tnear, tfar;
    int hit = intersectBox(eyeRay_o, eyeRay_d, boxMin, boxMax, &tnear, &tfar);
    if (!hit) 
    {
        if ((x < imageWidth) && (y < imageHeight)) 
        {
            // write output color
            uint i =(y * imageWidth) + x;
            pixels[i] = 0;
        }
        return;
    }

    if (tnear < 0.0f) 
        tnear = 0.0f;     // clamp to near plane

    // march along ray from back to front, accumulating color
    temp = (float4)(0.0f,0.0f,0.0f,0.0f);
    float t = tfar;
    
    int maxSteps = 500;
    float tstep = .01f;
    
    //initial intensity
    float I = 0;

    for(uint i=0; i<maxSteps; i++)
    {		
        float4 pos = eyeRay_o + eyeRay_d*t;
        pos = pos*0.5f+0.5f;    // map position to [0, 1] coordinates        
            
        float pX = pos.x * gridWidth;
        float pY = pos.y * gridHeight;
        float pZ = pos.z * gridDepth;        
        
        int index = indexOf(pX,pY,pZ,gridWidth, gridHeight, gridDepth);
        
        //Get the emission coefficient
        float j = volume[index];
        //Absorption coefficient
        float k = 0;        
        //The path length
        float ds = 1.f;//tstep;
        //Change in intensity
        float dI = 0;
        
        if(k == 0)
            dI = j*ds;
        else
        {
            float tau = k * ds;
            float S = j / k;
            float eTau = exp(-tau);
            dI = I * (eTau - 1) + S * (1 - eTau);
        }
        
        I += dI;    
       
        t -= tstep;
        
        if (t < tnear) 
           break;
    }
    
    temp.x = I;
    temp.y = I;
    temp.z = I;
    
    //Write to the image
    if ((x < imageWidth) && (y < imageHeight)) 
    {
        // write output color
        uint i =(y * imageWidth) + x;
        pixels[i] = rgbaFloatToInt(temp);
    } //end of if







//    int2 coord = (int2)(get_global_id(0), get_global_id(1));
//
//    float4 value = (float4)(0,0,0,0);
//
//            for(int z = 0; z < gridDepth; z++)
//            {
//                int index = indexOf(coord.x,coord.y,z,gridWidth,gridHeight,gridDepth);
//                float temp = volume[index];
//                value.x = value.x+temp;
//                value.y = value.y+temp;
//                value.z = value.z+temp;
//            }
//
//            int rp = value.x;
//            int gp = value.y;
//            int bp = value.z;
//
//            if(rp > 255)
//                rp = 255;
//            else if(rp < 0)
//                rp = 0;
//            if(gp > 255)
//                gp = 255;
//            else if(gp < 0)
//                gp = 0;
//            if(bp > 255)
//                bp = 255;
//            else if(bp < 0)
//                bp = 0;
//
//
//
//            int rgb = (rp << 16) | (gp <<  8) | (bp <<  0);
//
//            pixels[coord.y*imageWidth+coord.x] = rgb;
};


