
typedef struct { long a, b, c } random_state;


unsigned long random(random_state *r)
{
    unsigned long old = r->b;
    r->b = r->a * 1103515245 + 12345;
    r->a = (~old ^ (r->b >> 3)) - r->c++;
    return r->b;
}

float random_01(random_state *r)
{
    return (random(r) & 4294967295) / 4294967295.0f;
}

void seed_random(random_state *r, ulong seed)
{
    r->a = seed;
    r->b = 0;
    r->c = 362436;
}

/** 
* Calculates the 1D index of the 3D array with the given sizes.
*/
int indexOf(int x, int y, int z, int xSize, int ySize, int zSize)
{
    if(x < 0 || y < 0 || z < 0 || x >= xSize || y >= ySize || z >= zSize)
        return(-1);
        
    int index = z * (xSize*ySize) + y*xSize + x;

    return index;
} //end of indexOf function

/**
* intersect ray with a box
* http://www.siggraph.org/education/materials/HyperGraph/raytrace/rtinter3.htm
*/

int intersectBox(float4 r_o, float4 r_d, float4 boxmin, float4 boxmax, float *tnear, float *tfar)
{
    // compute intersection of ray with all six bbox planes
    float4 invR = (float4)(1.0f,1.0f,1.0f,1.0f) / r_d;
    float4 tbot = invR * (boxmin - r_o);
    float4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    float4 tmin = min(ttop, tbot);
    float4 tmax = max(ttop, tbot);

    // find the largest tmin and the smallest tmax
    float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

	*tnear = largest_tmin;
	*tfar = smallest_tmax;

	return smallest_tmax > largest_tmin;
} //end of intersectBox function

/**
* Returns the int RGBA value for the given float4
*/
uint rgbaFloatToInt(float4 rgba)
{
    rgba.x = clamp(rgba.x,0.0f,1.0f);  
    rgba.y = clamp(rgba.y,0.0f,1.0f);  
    rgba.z = clamp(rgba.z,0.0f,1.0f);  
    rgba.w = clamp(rgba.w,0.0f,1.0f);  
    return ((uint)(rgba.w*255.0f)<<24) | ((uint)(rgba.z*255.0f)<<16) | ((uint)(rgba.y*255.0f)<<8) | (uint)(rgba.x*255.0f);
} //end of rgbaFloatToInt function

/**
* Main kernel function
*/
__kernel void render(int imageWidth, int imageHeight, int gridWidth, int gridHeight, int gridDepth, __global float *matrix, __global float *jArray, __global float *kArray, __global float *output)
{
    uint x = get_global_id(0);
    uint y = get_global_id(1);
   
   //Write to the image
    if ((x >= imageWidth) && (y >= imageHeight)) 
    {
        return;
    } //end of if
   
    //Convert the indices to the range -1 -> + 1;
    float u = (x / (float) imageWidth)*2.0f-1.0f;
    float v = (y / (float) imageHeight)*2.0f-1.0f;

    float4 boxMin = (float4)(-1.0f, -1.0f, -1.0f,1.0f);
    float4 boxMax = (float4)(1.0f, 1.0f, 1.0f,1.0f);

    // calculate eye ray in world space
    float4 eyeRay_o = 0;
    float4 eyeRay_d = 0;

    eyeRay_o = (float4)(matrix[3], matrix[7], matrix[11], 1.0f);  
    //eyeRay_o = (float4)(u, v, -matrix[11], 1.0f);  
    

    float4 temp = normalize(((float4)(u, v, -2.0f,0.0f)));
    eyeRay_d.x = dot(temp, ((float4)(matrix[0],matrix[1],matrix[2],matrix[3])));
    eyeRay_d.y = dot(temp, ((float4)(matrix[4],matrix[5],matrix[6],matrix[7])));
    eyeRay_d.z = dot(temp, ((float4)(matrix[8],matrix[9],matrix[10],matrix[11])));
    eyeRay_d.w = 0.0f;
    
    //eyeRay_d.x = 0;
    //eyeRay_d.y = 0;
    //eyeRay_d.z = +1;
    

     // find intersection with box
    float tnear, tfar;
    int hit = intersectBox(eyeRay_o, eyeRay_d, boxMin, boxMax, &tnear, &tfar);
    if (!hit) 
    {
        if ((x < imageWidth) && (y < imageHeight)) 
        {
            // write output color
            uint i =(y * imageWidth) + x;
            output[i] = 0;
        }
        return;
    }

    if (tnear < 0.0f) 
        tnear = 0.0f;     // clamp to near plane

    // march along ray from back to front, accumulating color
    temp = (float4)(0.0f,0.0f,0.0f,0.0f);
    float t = tfar;

    int maxSteps = 5000;
    float tstep = .01f;

    //initial intensity
    float I = 0;

    for(uint i=0; i<maxSteps; i++)
    {
        float4 pos = eyeRay_o + eyeRay_d*t;

        pos = pos*0.5f+0.5f;    // map position to [0, 1] coordinates        

        float pX = pos.x * gridWidth/2.f;
        float pY = pos.y * gridHeight/2.f;
        float pZ = pos.z * gridDepth/2.f;        

        int index = indexOf(pX,pY,pZ,gridWidth, gridHeight, gridDepth);
        if(index < 0)
            continue;

        //Get the emission coefficient
        float j = jArray[index];
        //Absorption coefficient
        float k = kArray[index];        
        //The path length
        float ds = 1.f;//tstep;
        //Change in intensity
        float dI = 0;

        if(k == 0)
            dI = j*ds;
        else
        {
            float tau = k * ds;
            float S = j / k;
            float eTau = exp(-tau);
            dI = I * (eTau - 1) + S * (1 - eTau);
        }

        I += dI;    

        t -= tstep;

        if (t <= tnear) 
           break;
    } //end of for loop on i
       
    // write output value
    uint i =(y * imageWidth) + x;
    output[i] = I;
};


