
//#pragma OPENCL EXTENSION cl_khr_fp64 : enable

typedef struct { long a, b, c } random_state;


unsigned long random(random_state *r)
{
    unsigned long old = r->b;
    r->b = r->a * 1103515245 + 12345;
    r->a = (~old ^ (r->b >> 3)) - r->c++;
    return r->b;
}

float random_01(random_state *r)
{
    return (random(r) & 4294967295) / 4294967295.0f;
}

void seed_random(random_state *r, ulong seed)
{
    r->a = seed;
    r->b = 0;
    r->c = 362436;
}

/** 
* Calculates the 1D index of the 3D array with the given sizes.
*/
int indexOf(int x, int y, int z, int xSize, int ySize, int zSize)
{        
    if(z < 0 || y < 0 || x < 0 || x >= xSize || y >= ySize || z >= zSize)
        return(-1);

    int index = z * (xSize*ySize) + y*(xSize) + x;

    return index;
} //end of indexOf function

/** 
* Calculates the 1D index of the 3D array of velocities with the given sizes.
*/
int velIndexOf(int x, int y, int z, int xSize, int ySize, int zSize)
{        
    if(z < 0 || y < 0 || x < 0 || x >= xSize || y >= ySize || z >= zSize)
        return(-1);

    int index = z * (xSize*ySize*3) + y*(xSize*3) + x*3;

    return index;
} //end of indexOf function

/**
* intersect ray with a box
* http://www.siggraph.org/education/materials/HyperGraph/raytrace/rtinter3.htm
*/

int intersectBox(float4 r_o, float4 r_d, float4 boxmin, float4 boxmax, float *tnear, float *tfar)
{
    // compute intersection of ray with all six bbox planes
    float4 invR = (float4)(1.0f,1.0f,1.0f,1.0f) / r_d;
    float4 tbot = invR * (boxmin - r_o);
    float4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    float4 tmin = min(ttop, tbot);
    float4 tmax = max(ttop, tbot);

    // find the largest tmin and the smallest tmax
    float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

	*tnear = largest_tmin;
	*tfar = smallest_tmax;

	return smallest_tmax > largest_tmin;
} //end of intersectBox function



/**
* Returns the int RGBA value for the given float4
*/
uint rgbaFloatToInt(float4 rgba)
{
    rgba.x = clamp(rgba.x,0.0f,1.0f);  
    rgba.y = clamp(rgba.y,0.0f,1.0f);  
    rgba.z = clamp(rgba.z,0.0f,1.0f);  
    rgba.w = clamp(rgba.w,0.0f,1.0f);  
    return ((uint)(rgba.w*255.0f)<<24) | ((uint)(rgba.z*255.0f)<<16) | ((uint)(rgba.y*255.0f)<<8) | (uint)(rgba.x*255.0f);
} //end of rgbaFloatToInt function

//Interpolates the velocity around the position.  Pos is between 0 and 1
float4 interpVelocity(float4 pos, __global float *array, int n, int level)
{
    //Get the position in grid coordinates
    float xf = pos.x * (float)n;
    float yf = pos.y * (float)n;
    float zf = pos.z * (float)n;
    
    //get the index in the grid where the position is contained
    int x0 = round(xf);
    int y0 = round(yf);
    int z0 = round(zf);
   
    //The sum of all distances to adjacent cells
    float total = 0;    
    //Our velocity that we return
    float4 vel = {0,0,0,0};
    
    //Go through and calculate the distance of our point to each adjacent cell
    for(int x = x0-level; x <= x0+level; x++)
    {
        for(int y = y0-level; y <= y0+level; y++)
        {
            for(int z = z0-level; z <= z0+level; z++)
            {                
                float dx = xf - (float)x;
                float dy = yf - (float)y;
                float dz = zf - (float)z;
                
                //Store the distance
                float s = sqrt(dx*dx+dy*dy+dz*dz); 
                //Add to our total distance
                total = total + s;
                
                //Get the index into the velocity array
                int vIndex = velIndexOf(x,y,z,n,n,n);
                
                //Get the velocity components from the array
                float vx = array[vIndex+0];
                float vy = array[vIndex+1];
                float vz = array[vIndex+2];

                //Add to our total velocity
                vel.x = vel.x + s*vx;
                vel.y = vel.y + s*vy;
                vel.z = vel.z + s*vz;   
                    
            } //end of for loop on z
        } //end of for loop on y
    } //end of for loop on x
    
    //Now divide by the total to normalize
    
    if(total <= 0)
    {
        vel.x = 0;
        vel.y = 0;
        vel.z = 0;
    }
    
    else
    {
        vel.x = vel.x / total;
        vel.y = vel.y / total;
        vel.z = vel.z / total;    
    }
    
    return(vel);

} //end of interpolate function

//Interpolates the array around the position.  Pos is between 0 and 1
float interpPosition(float4 pos, __global float *array, int n, int level)
{
    //Get the position in grid coordinates
    float xf = pos.x * (float)n;
    float yf = pos.y * (float)n;
    float zf = pos.z * (float)n;
    
    //get the index in the grid where the position is contained
    int x0 = round(xf);
    int y0 = round(yf);
    int z0 = round(zf);
   
    //The sum of all distances to adjacent cells
    float total = 0;    
    //Our velocity that we return
    float value = 0;
    
    //Go through and calculate the distance of our point to each adjacent cell
    for(int x = x0-level; x <= x0+level; x++)
    {
        for(int y = y0-level; y <= y0+level; y++)
        {
            for(int z = z0-level; z <= z0+level; z++)
            {                
                float dx = xf - (float)x;
                float dy = yf - (float)y;
                float dz = zf - (float)z;
                
                //Store the distance
                float s = sqrt(dx*dx+dy*dy+dz*dz); 
                //Add to our total distance
                total = total + s;
                
                //Get the index into the velocity array
                int index = indexOf(x,y,z,n,n,n);
                
                //Get the velocity components from the array
                float v = array[index];

                //Add to our total velocity
                value = value + s*v;
                    
            } //end of for loop on z
        } //end of for loop on y
    } //end of for loop on x
    
    //Now divide by the total to normalize
    if(total <= 0)
        value = 0;
    else
        value = value / total;
    
    return(value);

} //end of interpolate function

/**
* Returns the pixel on the image where the velocity vZ should be
*/
int getVelocityPixel(int imageWidth, float velRange, float vZ)
{
    float vMin = 0.0f - velRange/2.f;

    int index = round((vZ-vMin) * ((float)imageWidth/velRange));

    return(index);
}

/**
* Main kernel function
*/
__kernel void pvrender(int imageWidth, int imageHeight, float velRange, int n, float range, __global float *matrix, __global float *jArray, __global float *kArray, __global float* vArray, __global float *output)
{
    uint x = get_global_id(0);
    uint y = get_global_id(1);
   
   //Write to the image
    if ((x >= imageWidth) && (y >= imageHeight)) 
    {
        return;
    } //end of if
   
    //Convert the indices to the range -1 -> + 1;
    float u = (x / (float) imageWidth)*2.0f-1.0f;
    float v = (y / (float) imageHeight)*2.0f-1.0f;

    float4 boxMin = (float4)(-1.0f, -1.0f, -1.0f,1.0f);
    float4 boxMax = (float4)(1.0f, 1.0f, 1.0f,1.0f);

    // calculate eye ray in world space
    float4 eyeRay_o = 0;
    float4 eyeRay_d = 0;

    eyeRay_o = (float4)(matrix[3], matrix[7], matrix[11], 1.0f);     

    float4 temp = {u, v, -2.0f,0.0f};
    eyeRay_d.x = dot(temp, ((float4)(matrix[0],matrix[1],matrix[2],matrix[3])));
    eyeRay_d.y = dot(temp, ((float4)(matrix[4],matrix[5],matrix[6],matrix[7])));
    eyeRay_d.z = dot(temp, ((float4)(matrix[8],matrix[9],matrix[10],matrix[11])));
    eyeRay_d.w = 0.0f;

     int ortho = 1;
    
    //Do extra processing if we want ortho
    if(ortho)
    {    
        //Shift the eye
        eyeRay_o.x = eyeRay_o.x + eyeRay_d.x;
        eyeRay_o.y = eyeRay_o.y + eyeRay_d.y;
        eyeRay_o.z = eyeRay_o.z + eyeRay_d.z;

        //Calculate the new direction vector  (make it parallel to screen)  
        eyeRay_d.x -= matrix[0]*u + matrix[1]*v;
        eyeRay_d.y -= matrix[4]*u + matrix[5]*v;
        eyeRay_d.z -= matrix[8]*u + matrix[9]*v;
        
    } //end of if
    
    //Make sure we are normalized!
    eyeRay_d = normalize(eyeRay_d);
  
     // find intersection with box
    float tnear, tfar;
    int hit = intersectBox(eyeRay_o, eyeRay_d, boxMin, boxMax, &tnear, &tfar);
    if (!hit) 
    {
        return;
    }
   
    if (tnear < 0.0f) 
        tnear = 0.0f;     // clamp to near plane

    // march along ray from back to front, accumulating color
    temp = (float4)(0.0f,0.0f,0.0f,0.0f);
    float t = tfar;

    int maxSteps = 500;
    float tstep = .01f;
    float ds = tstep * (range/2.f);

    //initial intensity
    float I = 0;
    
    //Scale to account for different areas of grid and pixels
    float scale = (range * range) / (imageWidth * imageHeight);
   
    for(uint i=0; i<maxSteps; i++)
    {
        float4 pos = eyeRay_o + eyeRay_d*t;

        pos = pos*0.5f+0.5f;    // map position to [0, 1] coordinates        

        int pX = round(pos.x * (float)(n-1));
        int pY = round(pos.y * (float)(n-1));
        int pZ = round(pos.z * (float)(n-1));        

        int index = indexOf(pX,pY,pZ,n,n,n);
        if(index < 0)
            break;
            
        //Get the emission coefficient
        float j = jArray[index];
        //Absorption coefficient
        float k = kArray[index];               
        //Change in intensity
        float dI = 0;

        if(k == 0)
            dI = j*ds;
        else
        {
            float tau = k * ds;
            float S = j / k;
            float eTau = exp(-tau);
            dI = I * (eTau - 1) + S * (1 - eTau);
        }
        
        //Get the interpolated velocity
        float4 vel = interpVelocity(pos, vArray, n, 0);
       
        //Get the velocity component projection onto the direction vector
        float vZ = dot(eyeRay_d, vel);

        //Get the pixel location on the PV diagram
        int xPix = getVelocityPixel(imageWidth, velRange, vZ);

        // write output value
        uint imageIndex =(y * imageWidth) + (xPix);
        output[imageIndex] += dI * scale;  

        t -= tstep;

        if (t < tnear) 
           break;
    } //end of for loop on i
    
};


